{- 

HOpenGL Tutorial
http://www.cin.ufpe.br/~hopengl/
Copyright (C) 2002  Andre Furtado <awbf@cin.ufpe.br>

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

This is a HOpenGL example.

-}

import GL
import GLUT
import IOExts
import UserInput
import Awbfbmp
foreign import shutdownHaskellAndExit :: Int -> IO ()

-- the player is defined by its x, y, z position and two angles indicating its horizontal and vertical rotation
type Player = (IORef GLdouble,IORef GLdouble,IORef GLdouble,IORef Int,IORef Int)

data Movement = Forwards | Backwards | SideStepRight | SideStepLeft
data Quad = Ceiling | Floor | SideWall | FrontBackWall

invColor :: Maybe InvisibleColorList
invColor = Just [(0,0,0)]

-- relates a Quad to its position in the textureList
texPosition :: Quad -> Int
texPosition Ceiling = 1
texPosition Floor = 2
texPosition SideWall = 3
texPosition FrontBackWall = 3

frameRate,winX,winY,winWidth,winHeight,rotSpeed, cX, cY :: Int
frameRate = 50
winX 	  = 100
winY 	  = 100
winWidth  = 800
winHeight = 600
rotSpeed  = 5
cX 	  = winWidth  `div` 2 -- center X position of window
cY        = winHeight `div` 2 -- center Y position of window

playerSpeed, playerHeight :: GLdouble
playerSpeed  = 1.0
playerHeight = 0.5

quadSize :: Float
quadSize     = 2.0

display :: Player -> [TextureName] -> DisplayAction
display (pX,pY,pZ,pH,pV) texs = do
	clear[ColorBufferBit] 
	playerX <- readIORef pX
	playerY <- readIORef pY
	playerZ <- readIORef pZ
	playerH <- readIORef pH
	playerV <- readIORef pV
	
	loadIdentity
	rotate ((fromIntegral playerH) - 270) (Vector3 0.0 1.0 (0.0 :: GLdouble))
--	rotate (playerV) (Vector3 1.0 0.0 (0.0 :: GLdouble)) --bugado!
	rotate (fromIntegral playerV) (Vector3 (sin (toRad playerH)) 0.0 (-(cos (toRad playerH))))
	translate (Vector3 (-playerX) (-playerY) (-playerZ))
	pushMatrix
	drawMap texs
	popMatrix
	drawPicture (getL 4 texs)
	flush
	swapBuffers
	
drawMap :: [TextureName]  -> IO ()
drawMap texs = do
	translate (Vector3 0.0 0.0 (-10.0 :: GLfloat)  )
	placeQuad Ceiling       texs   2.0
	placeQuad Floor         texs (-2.0)
	placeQuad SideWall      texs (-2.0)
	placeQuad SideWall      texs   2.0
	placeQuad FrontBackWall texs (-2.0)
	translate (Vector3 0.0 0.0 (2 * quadSize))
	placeQuad Ceiling       texs   2.0
	placeQuad Floor         texs (-2.0)
	placeQuad SideWall      texs (-2.0)
	placeQuad SideWall      texs   2.0
	translate (Vector3 0.0 0.0 (2 * quadSize))
	placeQuad Ceiling       texs   2.0
	placeQuad Floor         texs (-2.0)
	placeQuad SideWall      texs (-2.0)
	placeQuad SideWall      texs   2.0
	translate (Vector3 0.0 0.0 (2 * quadSize))
	placeQuad Ceiling       texs   2.0
	placeQuad Floor         texs (-2.0)
	placeQuad SideWall      texs (-2.0)
	placeQuad SideWall      texs   2.0
	placeQuad FrontBackWall texs   2.0
	
drawPicture :: TextureName -> IO ()
drawPicture tex = do
	translate (Vector3 0.0 0.0 (-6.0 :: GLfloat))
	enable Texture2d'
	bindTexture Texture2d tex
	color (Color3 1.0 1.0 1.0 :: Color3 GLfloat)
	beginEnd Quads $ do
      		texCoord2 0.0 0.0;  vertex3 2.0 (-1.0) (-1.5)
		texCoord2 0.0 1.0;  vertex3 2.0   1.0  (-1.5)
      		texCoord2 1.0 1.0;  vertex3 2.0   1.0    1.5
		texCoord2 1.0 0.0;  vertex3 2.0 (-1.0)   1.5
	disable Texture2d'
	
	
placeQuad :: Quad -> [TextureName] -> Float -> IO ()
placeQuad Ceiling texs n = do
	enable Texture2d'
	bindTexture Texture2d (getL 1 texs)
	color (Color3 1.0 1.0 1.0 :: Color3 GLfloat)
	beginEnd Quads $ do
      		texCoord2 0.0 0.0;  vertex3 (-quadSize) n (-quadSize)
		texCoord2 1.0 0.0;  vertex3 (-quadSize) n   quadSize
      		texCoord2 1.0 1.0;  vertex3   quadSize  n   quadSize
		texCoord2 0.0 1.0;  vertex3   quadSize  n (-quadSize)
	disable Texture2d'
placeQuad Floor texs n = do
	enable Texture2d'
	bindTexture Texture2d (getL 2 texs)
	color (Color3 1.0 1.0 1.0 :: Color3 GLfloat)
	beginEnd Quads $ do
      		texCoord2 0.0 0.0;  vertex3 (-quadSize) n (-quadSize)
		texCoord2 1.0 0.0;  vertex3   quadSize  n (-quadSize)
      		texCoord2 1.0 1.0;  vertex3   quadSize  n   quadSize
		texCoord2 0.0 1.0;  vertex3 (-quadSize) n   quadSize
	disable Texture2d'
placeQuad SideWall texs n = do
	enable Texture2d'
	bindTexture Texture2d (getL 3 texs)
	color (Color3 1.0 1.0 1.0 :: Color3 GLfloat)
	beginEnd Quads $ do
      		texCoord2 0.0 0.0;  vertex3 n (-quadSize) (-quadSize)
		texCoord2 0.0 1.0;  vertex3 n   quadSize  (-quadSize)
      		texCoord2 1.0 1.0;  vertex3 n   quadSize    quadSize
		texCoord2 1.0 0.0;  vertex3 n (-quadSize)   quadSize
	disable Texture2d'
placeQuad FrontBackWall texs n = do
	enable Texture2d'
	bindTexture Texture2d (getL 3 texs)
	color (Color3 1.0 1.0 1.0 :: Color3 GLfloat)
	beginEnd Quads $ do
      		texCoord2 0.0 0.0;  vertex3 (-quadSize) (-quadSize) n
		texCoord2 1.0 0.0;  vertex3   quadSize  (-quadSize) n
      		texCoord2 1.0 1.0;  vertex3   quadSize    quadSize  n
		texCoord2 0.0 1.0;  vertex3 (-quadSize)   quadSize  n
	disable Texture2d'
	
reshape :: ReshapeAction
reshape screenSize@(WindowSize w h) = do
   viewport (WindowPosition 0 0) screenSize
   matrixMode Projection
   loadIdentity
   let	near   = 1
	far    = 80
	fov    = 90
	ang    = (fov*pi)/(360 :: Double)
	top    = near / ( cos(ang) / sin(ang) )
	aspect = fromIntegral(w)/fromIntegral(h)
	right = top*aspect
   frustum (-right) right (-top) top near far
   matrixMode Modelview	

timer :: StillDownHandler -> (IORef Int, IORef Int) -> TimerAction
timer stillDown (oldMouseX, oldMouseY) = do
	stillDown
	postRedisplay
	warpPointer (WindowPosition cX cY)
	writeIORef oldMouseX cX
	writeIORef oldMouseY cY
	timerFunc frameRate (timer stillDown (oldMouseX, oldMouseY))
	
motion :: Player -> (IORef Int, IORef Int) -> MotionAction
motion (_,_,_,pH,pV) (oldMouseX,oldMouseY) (WindowPosition x y) = do
	oldX <- readIORef oldMouseX
	oldY <- readIORef oldMouseY
	let rX = (x - oldX) `div` rotSpeed
	    rY = (oldY - y) `div` rotSpeed
	
	updateIORef pH (\w -> (w + rX) `mod` 360)
	updateIORef pV (\w -> if (w + rY) > 90 then 90 else if (w + rY) < -90 then -90 else (w + rY))
	writeIORef oldMouseX x
	writeIORef oldMouseY y
	
createPlayer :: IO Player
createPlayer = do
	pX <- newIORef 0.0
	pY <- newIORef 0.0
	pZ <- newIORef 0.0
	pH <- newIORef 270 -- the player is looking to a point in the negative Z axis
	pV <- newIORef 0
	return (pX,pY,pZ,pH,pV)

-- Start of Aux Functions --

toRad :: Int -> GLdouble
toRad a = ((pi * (fromIntegral a))/180)

texStuff :: [TextureName] -> [AwbfBitmap] -> IO ()
texStuff [] _ = return ()
texStuff (t:ts) ((bmW,bmH,bmData):bms) = do
	bindTexture Texture2d t
	texImage2D Texture2d 0 GL.Rgba' bmW bmH 0 bmData
	texParameter Texture2d (TextureFilters Nearest Nearest)
	texStuff ts bms
texStuff _ _ = return ()

getL :: Int -> [t] -> t
getL _ [] = error "get list error: list empty"
getL 1 (a:_) = a
getL n as = getL (n-1) (tail as)

texCoord2 :: GLfloat -> GLfloat -> IO ()
texCoord2 u v = texCoord (TexCoord2 u v)

vertex3 :: GLfloat -> GLfloat -> GLfloat -> IO ()
vertex3 u v w = vertex (Vertex3 u v w)

-- End of Aux Functions --

awbfBindings :: Player -> IO (KeyBinder, StillDownHandler)
awbfBindings (pX,pY,pZ,pH,pV) = do
				
	(bindKey, stillDown) <- initUserInput
		
	bindKey (KeySpecial KeyLeft  ) StillDown (Just (updateIORef pH (\w -> (w - rotSpeed) `mod` 360)))
	bindKey (KeySpecial KeyRight ) StillDown (Just (updateIORef pH (\w -> (w + rotSpeed) `mod` 360)))
	bindKey (KeySpecial KeyHome  ) StillDown (Just (updateIORef pV (\w -> if (w + rotSpeed) >   90  then   90  else (w + rotSpeed) )))
	bindKey (KeySpecial KeyEnd   ) StillDown (Just (updateIORef pV (\w -> if (w - rotSpeed) < (-90) then (-90) else (w - rotSpeed) )))
	
	bindKey (KeySpecial KeyUp    ) StillDown (Just (move Forwards      pH (pX,pZ)))
	bindKey (KeySpecial KeyDown  ) StillDown (Just (move Backwards     pH (pX,pZ)))
	bindKey (KeyNormal  ','      ) StillDown (Just (move SideStepLeft  pH (pX,pZ)))
	bindKey (KeyNormal  '.'      ) StillDown (Just (move SideStepRight pH (pX,pZ)))
	
	bindKey (KeySpecial KeyPageUp     ) StillDown (Just (updateIORef pY (\w -> w + playerSpeed)))
	bindKey (KeySpecial KeyPageDown   ) StillDown (Just (updateIORef pY (\w -> w - playerSpeed)))
	
	bindKey (KeyNormal '\27'     ) Press     (Just (shutdownHaskellAndExit 1))

	return (bindKey, stillDown)
	

move :: Movement -> IORef Int -> (IORef GLdouble,IORef GLdouble) -> IO ()
move Forwards pH (pX,pZ) = do
	hAngle <- readIORef pH
	updateIORef pX (\w -> w + (playerSpeed * (cos (toRad hAngle)))) -- trigonometry!
	updateIORef pZ (\w -> w + (playerSpeed * (sin (toRad hAngle)))) -- trigonometry!
move Backwards pH (pX,pZ) = do
	hAngle <- readIORef pH
	updateIORef pX (\w -> w - (playerSpeed * (cos (toRad hAngle)))) -- trigonometry!
	updateIORef pZ (\w -> w - (playerSpeed * (sin (toRad hAngle)))) -- trigonometry!
move SideStepLeft pH (pX,pZ) = do
	hAngle <- readIORef pH
	updateIORef pX (\w -> w + ( playerSpeed * (sin (toRad hAngle)))) -- trigonometry!
	updateIORef pZ (\w -> w + (-playerSpeed * (cos (toRad hAngle)))) -- trigonometry!
move SideStepRight pH (pX,pZ) = do
	hAngle <- readIORef pH
	updateIORef pX (\w -> w + (-playerSpeed * (sin (toRad hAngle)))) -- trigonometry!
	updateIORef pZ (\w -> w + ( playerSpeed * (cos (toRad hAngle)))) -- trigonometry!

myInit :: IO ([TextureName])
myInit = do
	clearColor (Color4 0 0 0 0)
	clear [ColorBufferBit]
	shadeModel GL.Flat
	matrixMode Projection
  	loadIdentity
	matrixMode Modelview
	loadIdentity
	hint PerspectiveCorrectionHint Nicest
	texs  <- loadBitmapList (addInv ["ceiling.bmp","floor.bmp","wall.bmp","picture.bmp"])
	texBmList <- (genTextures (length texs ) )
	texStuff texBmList texs
	warpPointer (WindowPosition cX cY)
	setCursor CursorNone
	return (texBmList)

addInv :: [FilePath] -> [(FilePath,Maybe InvisibleColorList)]
addInv [] = []
addInv (a:as) = (a,invColor):(addInv as)

	
main :: IO ()
main = do
	GLUT.init Nothing
	player <- createPlayer
	oldMouseX <- newIORef cX
	oldMouseY <- newIORef cY
	createWindow "Haskell Doom - by Andre Furtado" (return ()) [ GLUT.Double, GLUT.Rgb ]
                (Just (WindowPosition winX     winY      ))
                (Just (WindowSize     winWidth winHeight ))
        texs <- myInit
        (bindKey, stillDown) <- awbfBindings player
	displayFunc (display player texs)
	timerFunc frameRate (timer stillDown (oldMouseX, oldMouseY))
	reshapeFunc (Just reshape)
	passiveMotionFunc (Just (motion player (oldMouseX, oldMouseY)))
	mainLoop  